import json
from http.server import BaseHTTPRequestHandler, HTTPServer

from .logs import logger


class ServerFinished(Exception):
    pass


class TestHandler(BaseHTTPRequestHandler):
    PORT = 7138
    def do_POST(self):
        content_length = int(self.headers["Content-Length"])
        data = json.loads(self.rfile.read(content_length))

        print(data["summary"])
        if data["finished"]:
            raise ServerFinished

        self.send_response(200)
        self.send_header("Content-Type", "text/plain")
        self.end_headers()
        self.wfile.write("ok")

    def log_message(self, *args, **kwargs):
        pass


class TestServer:
    ADDRESS = f"http://127.0.0.1:{TestHandler.PORT}"

    @classmethod
    def serve(cls):
        server = HTTPServer(("", TestHandler.PORT), TestHandler)
        logger.debug(f"Listening to localhost:{TestHandler.PORT}")
        server.serve_forever()




def default_kafka() -> dict:
    return {
        "image": "confluentinc/confluent-local:7.4.1",
        "hostname": "kafka",
        "container_name": "kafka",
        "ports": ["9092:9092"],
        "environment": {
            "KAFKA_ADVERTISED_LISTENERS": "PLAINTEXT://kafka:29092,PLAINTEXT_HOST://localhost:9092",
            "KAFKA_CONTROLLER_QUORUM_VOTERS": "1@kafka:29093",
            "KAFKA_LISTENERS": "PLAINTEXT://kafka:29092,CONTROLLER://kafka:29093,PLAINTEXT_HOST://0.0.0.0:9092",
        },
        "healthcheck": {
            "test": ["CMD", "sleep", "1"],
            "retries": 0,
            "start_period": "15s",
            "start_interval": "0s"
        }
    }


def default_producer(
    config_dir: str,
    message_dir: str = "/var/kafka_messages",
    depends_on: dict|None = None
):
    if depends_on is None:
        depends_on = {}

    name = "at_prod"
    return {
        "hostname": name,
        "container_name": name,
        "image": "registry.apps.eo4eu.eu/eo4eu/eo4eu-provision-handler/utils/autotest-utils/autotest-producer:latest",
        "environment": [
            "BOOTSTRAP_SERVERS=kafka:29092",
            "GROUP_ID=at_prod",
            f"MESSAGE_DIR={message_dir}/*"
        ],
        "depends_on": {
            "at_serv": {"condition": "service_healthy"},
        } | depends_on,
        "volumes": [
            f"{config_dir}:{message_dir}"
        ]
    }


def default_service(config_dir: str):
    name = "at_serv"
    return {
        "hostname": name,
        "container_name": name,
        "image": "registry.apps.eo4eu.eu/eo4eu/eo4eu-provision-handler/utils/autotest-utils/autotest-service:latest",
        "environment": [
            "BOOTSTRAP_SERVERS=kafka:29092",
            "GROUP_ID=at_serv",
            "CONFIG_FILE=/var/config/autotest_service.json",
        ],
        "depends_on": {
            "kafka": {"condition": "service_healthy"},
        },
        "volumes": [
            f"{config_dir}:/var/config",
        ]
    }


def default_canary():
    return {
        "version": "0.1.4",
        "options": {
            "interval_sec": 2,
            "max_misses": 5,
            "initial_max_misses": 10,
        },
        "check": {
            "interval": "2s",
            "start_period": "0s",
            "start_interval": "0s"
        },
        "gitlab": {
            "username": "GITLAB_USERNAME",
            "token": "GITLAB_TOKEN",
        }
    }



from pydantic import BaseModel
from pathlib import Path
from typing import Dict
import tomllib
import json


def read_version(version_filepath: Path) -> str:
    version_txt = version_filepath.read_text()
    if version_filepath.name == "pyproject.toml":
        data = tomllib.loads(version_txt)
        return data["project"]["version"]
    if version_filepath.suffix == ".json":
        data = json.loads(version_txt)
        return data["version"]
    else:
        return version_txt


class ImageLockInfoV1(BaseModel):
    version: str
    hash: str


class LockFileV1(BaseModel):
    version: str
    images: Dict[str,ImageLockInfoV1]



from typing import Optional, List, Dict
from pydantic import BaseModel
from pathlib import Path
import yaml
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

from .builder import ImageBuilder
from .logs import logger



class RegistryConfigV1(BaseModel):
    reg: str
    url: str


class FilterConfigV1(BaseModel):
    tag:        Optional[List[str]] = None
    registry:   Optional[List[str]] = None
    dockerfile: Optional[List[str]] = None


class BuildConfigV1(BaseModel):
    tag:        Optional[str] = "latest"
    registry:   str
    dockerfile: Optional[str] = "Dockerfile"


class BuilderConfigV1(BaseModel):
    project_root: str
    version_file: str
    filters:      Optional[dict[str,FilterConfigV1]] = None
    registries:   Dict[str,RegistryConfigV1]
    builds:       List[BuildConfigV1]


def get_builder_config(version: str, input: dict) -> BuilderConfigV1:
    if version == "1":
        result = BuilderConfigV1.model_validate(input)
        if result.filters is None:
            result.filters = {}
        if "_default" not in result.filters:
            result.filters["_default"] = FilterConfigV1()
        return result
    else:
        logger.warning(f"Invalid version \"{version}\" of config file. Defaulting to \"1\"")
        return get_builder_config("1", input)

def parse_recipe_file(filename: str) -> dict[str,ImageBuilder]:
    filter_names = ["_default"]
    if ":" in filename:
        filename, filter_str = filename.split(":")
        filter_names = [filter_str]
        if "," in filter_str:
            filter_names = filter_str.split(",")

    logger.debug(f"Reading recipe file {filename} with filters {filter_names} enabled")
    try:
        data = yaml.load(Path(filename).read_text(), Loader = Loader)
    except Exception as e:
        logger.error(f"Failed to parse recipe file {filename}")
        raise e

    version = "1"
    if "version" in data:
        version = data["version"]
        del data["version"]
    else:
        logger.info(f"Version not specified, defaulting to \"1\"")

    result = {}
    for key, config_dict in data.items():
        try:
            config = get_builder_config(version, config_dict)
            local_filter_names = [
                filter_name for filter_name in filter_names
                if filter_name in config.filters
            ]
            if len(local_filter_names) == 0:
                local_filter_names = ["_default"]

            logger.debug(f"Parsing config for {key} with filters {local_filter_names}")
            builder = ImageBuilder.from_config_v1(config)
            for filter_name in local_filter_names:
                builder = builder.filter(config.filters[filter_name].model_dump())

            result[key] = builder
        except Exception as e:
            logger.error(f"Failed to configure builder {key}: {e}")
            continue

    return result



import subprocess
from pathlib import Path
from importlib import import_module
from importlib.util import spec_from_file_location, module_from_spec

from .server import TestServer, ServerFinished
from .config import TestConfig
from .logs import logger


def import_config(args) -> TestConfig:
    config_path = Path.cwd().joinpath(args.file)
    if not config_path.is_file():
        raise FileNotFoundError(f"Failed to find config file {args.file}")

    spec = spec_from_file_location(
        config_path.stem.replace("-", "_"),
        config_path
    )
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    config = module.CONFIG
    config.set_top_dir(config_path.parent)
    return config


def build(args, config: TestConfig):
    """eo-test build|b [-f|--file?=eo-test.py]
    Builds the docker compose file and all related dependencies
    needed for the mock tests, as specified in --file.
    examples:
        eo-test build
            builds using ${PWD}/eo-test.py
        eo-test b -f other-eo-test.py
            builds using ${PWD}/other-eo-test.py
    """
    logger.debug(f"building {config.compose_file()}...")
    config.build_docker_compose()


def run(args, config: TestConfig):
    """eo-test run|r [-f|--file?=eo-test.py]
    (Optionally) builds and runs the mock tests
    examples:
        eo-test run
            runs using ${PWD}/eo-test.py
        eo-test r -f other-eo-test.py
            runs using ${PWD}/other-eo-test.py
        eo-test r --no-rebuild
            runs using ${PWD}/eo-test.py without rebuilding the configuration
    """
    # FIXME: connect to localhost server from inside
    # docker compose
    # config.set_forward_address(TestServer.ADDRESS)
    build(args, config)

    docker_compose_path = config.compose_file()
    subprocess.run([
        "docker", "compose",
        "-f", docker_compose_path,
        "up",
        "-d"
    ])
    subprocess.run([
        "docker", "compose",
        "-f", docker_compose_path,
        "logs",
        "at_serv",
        "--no-log-prefix",
        "--follow",
        "--tail", "10"
    ])
    # try:
    #     TestServer.serve()
    # except ServerFinished:
    #     return
    # except Exception as e:
    #     logger.error(f"Unexpected error: {e}")


def clean(args, config: TestConfig):
    """eo-test clean|c [-f|--file?=eo-test.py]
    Cleans up test artifacts and shuts down running containers
    examples:
        eo-test clean
            cleans using ${PWD}/eo-test.py
        eo-test c -f other-eo-test.py
            cleans using ${PWD}/other-eo-test.py
    """
    config.cleanup()

    docker_compose_path = config.compose_file()
    subprocess.run([
        "docker", "compose",
        "-f", docker_compose_path,
        "down",
    ])


HELP_ALL = """eo-test: A command-line utility for running tests on components.
         It accepts a config file written in Python which defines
         "mock" tests; these are simulations of the real component
         running conditions, using a local docker compose file
         which sets up a kafka server and other dependencies.
Usage: eo-test [command] [options...]
Commands: help  | h
          build | b
          run   | r
          clean | c
Use eo-test help [command] to get a help blurb for individual commands
"""

def help(cmd_dict, cmd_name_dict, *commands: str) -> str:
    """eo-test help|h [command...]
    Prints help messages
    """
    if len(commands) == 0:
        return HELP_ALL

    for command in commands:
        return cmd_dict[cmd_name_dict[command]].__doc__


cmd_dict = {
    "help":  help,
    "build": build,
    "run":   run,
    "clean": clean,
}

cmd_name_dict = {
    "help":  "help",
    "build": "build",
    "run":   "run",
    "clean": "clean",
    "h":     "help",
    "b":     "build",
    "r":     "run",
    "c":     "clean",
}



import json
from typing import Any


class KafkaMessage:
    def __init__(self, topic: str, value: Any):
        self._topic = topic
        self._value = value

    def topic(self) -> str:
        return self._topic

    def value(self) -> Any:
        return self._value

    def json(self) -> str:
        return json.dumps(self.value())



import os
import sys
import copy
import stat
import json
import shutil
import requests
from pathlib import Path
from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

from .logs import logger
from .recipe import Recipe
from .server import TestServer
from .message import KafkaMessage
from .defaults import (
    default_kafka,
    default_producer,
    default_service,
    default_canary,
)

class TestConfig:
    def __init__(
        self,
        recipes: list[Recipe],
        working_dir: str = "",
        compose_file: str = "docker-compose.yaml",
        **kwargs
    ):
        self._top_dir = Path.cwd()
        self._working_dir = working_dir
        self._compose_file = compose_file
        self._recipes = recipes
        self._attrs = kwargs
        self._forward_to = None

    def set_forward_address(self, address: str):
        self._forward_to = address

    def set_top_dir(self, top_dir: Path):
        self._top_dir = top_dir

    def recipes(self) -> list[Recipe]:
        return self._recipes

    def working_dir(self) -> Path:
        return self._top_dir.joinpath(self._working_dir)

    def compose_file(self) -> Path:
        return self.working_dir().joinpath(self._compose_file)

    def local_cfg_dir(self) -> Path:
        return self.working_dir().joinpath(Recipe.LOCAL_CFG_DIR)

    def to_dict(self, script_path: str|None = None) -> dict:
        self.local_cfg_dir().mkdir(parents = True, exist_ok = True)

        result = copy.deepcopy(self._attrs)
        result["services"] = result.get("services", {})
        result["services"]["kafka"] = default_kafka() | result["services"].get("kafka", {})

        canary_config = self._add_canary(result.get("canary", True))
        self._add_producer(result, canary_config)
        self._add_service(result, canary_config)
        self._add_recipes(result, canary_config)
        return result

    def to_yaml(self) -> str:
        return dump(self.to_dict(), Dumper = Dumper)

    def build_docker_compose(self):
        self.compose_file().write_text(self.to_yaml())

    def cleanup(self):
        if self.local_cfg_dir().exists():
            shutil.rmtree(self.local_cfg_dir())

    def _download_canary(self, canary_config: dict) -> bool:
        username, token = "", ""
        try:
            username = os.environ[canary_config["gitlab"]["username"]]
            token = os.environ[canary_config["gitlab"]["token"]]
        except Exception as e:
            logger.warning(f"Unable to get gitlab credentials "
                           f"in order to download autotest-canary: {e}")
            return False

        version = ""
        if "version" in canary_config:
            version = canary_config["version"]
        else:
            version = default_canary()["version"]
            logger.warning(f"Version for autotest-canary not found, "
                           f"defaulting to {version}")

        logger.debug("Downloading autotest-canary...")
        response = requests.get(
            "https://git.apps.eo4eu.eu/api/v4/projects/154/packages/generic"
            f"/autotest-canary/{version}/autotest-canary",
            auth = (username, token)
        )

        if response.status_code != 200:
            logger.warning(f"Failed to download autotest-canary: "
                           f"HTTP status code {response.status_code}")
            return False

        out_path = self.local_cfg_dir().joinpath("autotest-canary")
        out_path.write_bytes(response.content)
        out_path.chmod(out_path.stat().st_mode | stat.S_IEXEC)
        return True

    def _add_canary(self, canary_config: dict|bool) -> dict|bool:
        if canary_config == False:
            return False
        if canary_config == True:
            canary_config = default_canary()

        downloaded = self._download_canary(canary_config)
        if downloaded:
            return canary_config
        else:
            logger.warning(f"Unable to download autotest-canary, disabling feature...")
            return False

    def _add_producer(self, result: dict, canary_config: dict|bool):
        # write the input messages
        at_prod_dir = self.local_cfg_dir().joinpath("at_prod")
        if at_prod_dir.exists():
            shutil.rmtree(at_prod_dir)
        at_prod_dir.mkdir(parents = True, exist_ok = True)
        for recipe in self._recipes:
            for topic, message in recipe.inputs().items():
                at_prod_dir.joinpath(topic).write_text(json.dumps(message))

        condition = "service_healthy" if canary_config != False else "service_started"
        result["services"]["at_prod"] = default_producer(
            config_dir = self.local_cfg_dir().joinpath("at_prod"),
            depends_on = {
                recipe.name(): {"condition": condition}
                for recipe in self._recipes
            }
        ) | result["services"].get("at_prod", {})

    def _add_service(self, result: dict, canary_config: dict|bool):
        at_serv_dir = self.local_cfg_dir().joinpath("at_serv")
        if at_serv_dir.exists():
            shutil.rmtree(at_serv_dir)
        at_serv_dir.mkdir(parents = True, exist_ok = True)
        at_serv_dir.joinpath("autotest_service.json").write_text(
            json.dumps(self._get_service_config(canary_config))
        )

        result["services"]["at_serv"] = default_service(
            config_dir = self.local_cfg_dir().joinpath("at_serv"),
        ) | result["services"].get("at_serv", {})

    def _add_recipes(self, result: dict, canary_config: dict|bool):
        for recipe in self._recipes:
            recipe_name = recipe.name()
            recipe_conf = recipe.to_dict(
                working_dir = self.working_dir(),
                canary_config = canary_config
            )
            if recipe_name in result["services"]:
                result["services"][recipe_name] = recipe_conf | result["services"][recipe_name]
            else:
                result["services"][recipe_name] = recipe_conf

            for top_level, config in recipe.also_add().items():
                if top_level in result:
                    result[top_level] |= config
                else:
                    result[top_level] = config

    def _get_service_config(self, canary_config: dict|bool) -> dict:
        has_canary = canary_config != False
        create_topics = set()
        recipes = {}
        for recipe in self._recipes:
            recipe_name = recipe.name()
            recipes[recipe_name] = {"match": {}}
            if has_canary:
                canary_topic = f"{recipe_name}-canary_"
                recipes[recipe_name]["canary"] = {
                    "topic": canary_topic,
                    **canary_config["options"],
                }
                create_topics.add(canary_topic)

            for topic, message in recipe.outputs().items():
                create_topics.add(topic)
                recipes[recipe_name]["match"][topic] = [message]
            for topic in recipe.inputs().keys():
                create_topics.add(topic)
        result = {
            "create_topics": list(create_topics),
            "recipes": recipes,
        }
        if self._forward_to is not None:
            result["forward_to"] = self._forward_to

        return result



import os
from pathlib import Path

from .message import KafkaMessage
from .defaults import default_canary


class Recipe:
    LOCAL_CFG_DIR = ".config"
    CONTAINER_CFG_DIR = Path("/")

    def __init__(
        self,
        name: str,
        image: str,
        configs: dict[str,str]|None = None,
        secrets: dict[str,str]|None = None,
        env: dict[str,str]|None = None,
        inputs: list[KafkaMessage]|None = None,
        outputs: list[KafkaMessage]|None = None,
        depends_on: dict[str,dict[str,str]]|None = None,
        also_add: dict|None = None,
        **kwargs
    ):
        if configs is None:
            configs = {}
        if secrets is None:
            secrets = {}
        if env is None:
            env = {}
        if inputs is None:
            inputs = []
        if outputs is None:
            outputs = []
        if depends_on is None:
            depends_on = {}
        if also_add is None:
            also_add = {}
        if "at_serv" not in depends_on:
            depends_on["at_serv"] = {
                "condition": "service_healthy"
            }

        self._name = name
        self._image = image
        self._configs = configs
        self._secrets = secrets
        self._env = env
        self._inputs = inputs
        self._outputs = outputs
        self._depends_on = depends_on
        self._also_add = also_add
        self._kwargs = kwargs

    def name(self) -> str:
        return self._name

    def inputs(self) -> list[KafkaMessage]:
        return self._inputs

    def outputs(self) -> list[KafkaMessage]:
        return self._outputs

    def also_add(self) -> dict:
        return self._also_add

    def to_dict(self, working_dir: Path, canary_config: dict|bool = True) -> dict:
        cfg_dir = working_dir.joinpath(self.LOCAL_CFG_DIR, self._name)
        for mountpoint, content in self._configs.items():
            path = cfg_dir.joinpath(mountpoint)
            if not path.parent.is_dir():
                path.parent.mkdir(parents = True)
            path.write_text(content)

        result = {
            "image": self._image,
            "container_name": self._name,
            "environment": [
                f"{key}={val}"
                for key, val in self._env.items()
            ],
            "secrets": [
                {"source": secret_name, "target": str(Path("/").joinpath(mountpoint))}
                for mountpoint, secret_name in self._secrets.items()
            ],
            "volumes": [
                ":".join([
                    str(cfg_dir.joinpath(mountpoint)),
                    str(self.CONTAINER_CFG_DIR.joinpath(mountpoint))
                ])
                for mountpoint in self._configs.keys()
            ],
            "depends_on": self._depends_on,
            **self._kwargs
        }

        if canary_config == False:
            return result

        if canary_config == True:
            canary_config = {}
        canary_config = default_canary() | canary_config
        result["volumes"].append(":".join([
            str(cfg_dir.parent.joinpath("autotest-canary")),
            "/bin/autotest-canary"
        ]))
        result["healthcheck"] = {
            "test": ["CMD", "/bin/autotest-canary"],
            **canary_config["check"]
        }

        return result



from argparse import ArgumentParser
from pathlib import Path
from pprint import pprint
import logging

from .logs import logger
from .builder import ImageBuilder
from .recipe import parse_recipe_file


def run():
    parser = ArgumentParser(add_help = False)
    parser.add_argument("commands", type = str, nargs = "*")
    parser.add_argument("--debug", "-d", action = "store_true")
    parser.add_argument("--lockfile-dir", default = ".cons.lock")

    args = parser.parse_args()
    if args.debug:
        logger.setLevel(logging.DEBUG)

    lockfile_dir = Path(args.lockfile_dir)
    if not lockfile_dir.is_absolute():
        lockfile_dir = Path.cwd().joinpath(lockfile_dir)

    commands = args.commands
    if len(commands) < 1:
        logger.error("Need at least 1 command")
        return

    command, rest = commands[0], commands[1:]
    action_dict = {
        "build": False,
        "deploy": True,
    }
    if command in action_dict:
        if len(rest) < 1:
            logger.critical(f"Command \"{command}\" needs at least 1 argument")
            return

        push = action_dict[command]
        for filename in rest:
            builders = parse_recipe_file(filename)
            for key, builder in builders.items():
                # print(key)
                # print(builder)
                logger.info(f"Building {key}: version {builder.version}...")
                builder.build(push)
    else:
        logger.critical(f"Command \"{command}\" not recognized.")



from argparse import ArgumentParser
from pathlib import Path
from pprint import pprint
import traceback
import logging

from .commands import import_config, cmd_dict, cmd_name_dict
from .logs import logger

command_dict = {
    "build": False,
    "deploy": True,
}

def run():
    parser = ArgumentParser(add_help = False)
    parser.add_argument("commands", type = str, nargs = "*")
    parser.add_argument("--file", "-f", type = str, default = "eo-test.py")
    parser.add_argument("--debug", "-d", action = "store_true")

    args = parser.parse_args()
    if args.debug:
        logger.setLevel(logging.DEBUG)

    commands = args.commands
    if len(commands) < 1:
        logger.error("Need at least 1 command")
        return

    command, rest = commands[0], commands[1:]
    try:
        command = cmd_name_dict[command]
    except KeyError:
        logger.critical(f"Command \"{command}\" not recognized. "
                        "Use eo-test help for a list of available commands")
        return

    cmd = cmd_dict[command]
    if command == "help":
        print(cmd(cmd_dict, cmd_name_dict, *rest))
        return

    config = None
    try:
        config = import_config(args)
    except Exception as e:
        logger.critical(f"Failed to read config: {e}")
        if args.debug:
            logger.critical(traceback.fmt_exc())
        return

    try:
        cmd(args, config)
    except Exception as e:
        logger.critical(f"Error: {e}")
        if args.debug:
            logger.critical(traceback.format_exc())



from typing import Self
from pathlib import Path
from pprint import pformat
import subprocess
import yaml
import copy

from .logs import logger
from .lockfile import read_version


class ImageBuild:
    def __init__(
        self,
        login_cmd: list[str],
        build_cmd: list[str],
        push_cmd: list[str],
        features: dict,
        image: str
    ):
        self.login_cmd = login_cmd
        self.build_cmd = build_cmd
        self.push_cmd = push_cmd
        self.features = features
        self.image = image
        self._output = {}

    @classmethod
    def default(
        cls,
        project_root: str,
        dockerfile: str|None,
        registry_name: str,
        registry_url: str,
        tag: str,
        features: dict
    ):
        image = f"{registry_url}:{tag}"
        if dockerfile is None:
            build_cmd = ["docker", "build", "-t", image, project_root]
        else:
            build_cmd = ["docker", "build", "-f", dockerfile, "-t", image, project_root]
        return ImageBuild(
            login_cmd = ["docker", "login", registry_name],
            build_cmd = build_cmd,
            push_cmd = ["docker", "push", registry_url],
            features = features,
            image = image
        )

    def _exec(self, name: str, command: list[str]) -> Self:
        logger.info(" ".join(command))
        output = subprocess.run(command, capture_output = True)
        stdout = output.stdout.decode("utf-8")
        stderr = output.stderr.decode("utf-8")
        stdout = f"{stdout}\n" if stdout != "" else stdout
        total = stdout + stderr
        self._output[name] = total
        if output.returncode != 0:
            raise OSError(
                f"Command returned exit status {output.returncode}:\n{total}"
            )
        else:
            logger.debug(total)

        return self

    def login(self) -> Self:
        return self._exec("login", self.login_cmd)

    def build(self) -> Self:
        return self._exec("build", self.build_cmd)

    def push(self) -> Self:
        return self._exec("push", self.push_cmd)

    # def get_hash(self) -> str|None:
    #     if "build" in self._output:
    #         build_output = self._output["build"].split("\n")
    #         for line in reversed(build_output):
    #             if "writing image" not in build_output:
    #                 continue
    #             a
    #     return None

    def __repr__(self) -> str:
        return pformat(self.features)


class ImageBuilder:
    def __init__(self, version: str, builds: list[ImageBuild]):
        self.version = version
        self.builds = builds

    @classmethod
    def from_config_v1(self, input) -> Self:
        try:
            builds = []
            project_root = Path.cwd().joinpath(input.project_root).resolve()
            version_file = project_root.joinpath(input.version_file)
            registries = input.registries

            for build in input.builds:
                registry = registries[build.registry]
                builds.append(ImageBuild.default(
                    project_root = str(project_root),
                    dockerfile = None if build.dockerfile == "Dockerfile" else str(
                        project_root.joinpath(build.dockerfile)
                    ),
                    registry_name = registry.reg,
                    registry_url = registry.url,
                    tag = build.tag,
                    features = build.model_dump()
                ))

            return ImageBuilder(
                version = read_version(version_file),
                builds = builds
            )
        except Exception as e:
            logger.error(f"Failed to configure ImageBuilder: {e}")
            return None

    def with_builds(self, builds: list[ImageBuild]) -> Self:
        return ImageBuilder(
            version = self.version,
            builds = builds
        )

    def filter(self, features: dict[str,list[str]]) -> Self:
        filtered_builds = []
        for build in self.builds:
            if all([
                values is None or len(values) == 0 or build.features[feature] in values
                for feature, values in features.items()
            ]):
                filtered_builds.append(build)

        return self.with_builds(filtered_builds)

    def build(self, push: bool = False):
        for build in self.builds:
            try:
                if push:
                    build.login().build().push()
                else:
                    build.build()
            except Exception as e:
                logger.error(f"Deployment failed: {e}")

    def __repr__(self) -> str:
        return "\n".join([
            f"version: {self.version}",
            pformat(self.builds)
        ])
